import wsgiref.handlers
import logging

import base64
import hmac
import sha 
import time

import urllib 
from google.appengine.api import urlfetch, users
from google.appengine.ext import webapp, db
from google.appengine.ext.webapp.util import login_required

from django.utils import simplejson as json

# the SMSUser data Model
class SMSUser(db.Model):
    # two properties... the google user account and the mobile phone number
    account = db.UserProperty()
    mobile_number = db.PhoneNumberProperty()

    # add a record for this mobile number
    def authenticate(self, mobile_number):
        u = SMSUser(mobile_number=mobile_number)
        u.put()

    # is a specific mobile_number already authenticated?
    def is_authenticated(self, mobile_number):
        sms_user = SMSUser.gql('WHERE mobile_number = :1', mobile_number).get()
        if sms_user: 
            return True
        else:
            return False


class SMSService(object):
    def __init__(self, short_code, keyword, api_key=None, api_secret=None):
        self._short_code = short_code
        self._keyword = keyword
        self._api_key = api_key
        self._api_secret = api_secret


    @property 
    def short_code(self):
        return self._short_code

    def send(self, number, body):
        # override this method
        logging.info("I should send a message to %s" % number)

    # Is the current user authenticated?
    def is_authenticated(self, mobile_number):
        # pass this up the chain
        return SMSUser().is_authenticated(mobile_number)

    # pass this up the chain
    def authenticate(self, mobile_number):
        return SMSUser().authenticate(mobile_number)

    # should we authenticate a given mobile number?    
    def should_authenticate(self, mobile_number, event, body):
        if mobile_number:
            return True

    # if the user isn't authenticated (and needs to be) send this message.
    def signup_message(self):
        # Let's assume the obvious.. subclasses can override this
        return "Text '%s JOIN' to %s sign up for this service." % (self._keyword, 
                                                                   self._short_code)

class SMSBaseHandler(webapp.RequestHandler):
    
    def __init__(self):
        self.user = SMSUser()
        # a dictionary filled with the supported services
        self._supported_services = {
            'textmark': Textmark,
            'zeep': Zeep,
            }

        # the service being used
        self._service = None

    # the controller will need to tell us which class to use
    def set_service_type(self, service_type):
        if service_type in self._supported_services:
            self._service = self._supported_services[service_type]()
        else:
            raise Exception('type', 'Invalid SMS Service Type')

    # a convenience property allowing easy access to the resulting class
    @property 
    def service(self):
        if self._service:
            return self._service
        # no service is defined, throw an exception
        else:
            raise Exception('noservice', 'No service type set.')




class SMSIncomingHandler(SMSBaseHandler):
    # handle the GET requests
    def get(self, service_type):
        self.respond(service_type)

    # handle the POST requests
    def post(self, service_type):
        self.respond(service_type)

    # all of the HTTP requests end up handled by this method
    def respond(self, service_type):

        # setup the service type
        self.set_service_type(service_type)

        # start the response
        self.response.headers['Content-Type'] = 'text/plain'

        # prepare some HTTP parameters as variables
        uid = self.request.get('uid')
        event = self.request.get('event')
        body = self.request.get('body')

        # are we authenticated?
        if uid and not self.service.is_authenticated(uid):
            # should we authenticate?
            logging.info("Not authenticated, should we? (uid: %s)" % uid)
            if self.service.should_authenticate(uid, event, body):
                # dooo it.
                logging.info("Authenticating: %s " % uid)
                self.service.authenticate(uid)
            else:
                logging.info("No reason to authenticate uid: %s" % uid)
        elif len(uid):
            logging.info("Previously authenticated uid: %s" % uid)
        else:
            logging.info("No UID, not authenticating")

        # respond to the request
        self.response.headers['Content-Type'] = 'text/plain'
        self.response.out.write("Hello there via %s!" % self.service.short_code)


class Textmark(SMSService):
    def __init__(self):
        self._auth_user = '123456789'  # USER DEFINED
        self._auth_pass = '1234'  # USER DEFINED
        SMSService.__init__(self, 
                            '41411',   # the short code
                            'realtime',
                            api_key='12345',  # USER DEFINED
                            )
 
    # override the send method from SMSService
    def send(self, number, body):
        # The Send URL
        url = "http://dev1.api2.textmarks.com/GroupLeader/send_one_message/"

        # setup the http arguments
        args = {}
        args['to'] = number             
        args['msg'] = body
        args['tm'] = self._keyword
        args['api_key'] = self._api_key       
        args['auth_user'] = self._auth_user
        args['auth_pass'] = self._auth_pass   

        # make the HTTP API call to send the message
        result = urlfetch.fetch(url=url,
                                payload=urllib.urlencode(args),
                                method=urlfetch.POST)
        logging.info(result.content)


class Zeep(SMSService):
    def __init__(self):
        SMSService.__init__(self, 
                            '88147',   # the short code
                            'realtime',
                            api_key='12345',
                            api_secret='12345')
    
    def send(self, number, body):
        # the payload to send (and for the signature)
        payload = "user_id=%s&body=%s" % (number, urllib.quote(body))

        # the timestamp needed for the signature
        ts = time.strftime('%a, %d %b %Y %H:%M:%S GMT',time.gmtime())

        # the main part of the signature
        sig_content = "%s%s%s" % (self._api_key, ts, payload)
        
        # start building the signature with the api secret
        sig = hmac.new(self._api_secret, digestmod=sha)

        # add the main content to the signature hash
        sig.update(sig_content)

        # generate and encode the hash
        usable_sig = base64.encodestring(sig.digest()).strip()

        # setup the authorization headers
        headers = {'Date': ts,
                   'Authorization': 'Zeep %s:%s' % (self._api_key, usable_sig)
                   }
        
        # the send API url
        url = "https://api.zeepmobile.com/messaging/2008-07-14/send_message"

        # make the request
        result = urlfetch.fetch(url=url,
                                payload=payload,
                                method=urlfetch.POST,
                                headers=headers
                                )

        logging.info(headers)
        logging.info(result.content)
        
    def should_authenticate(self, mobile_number, event, body):
        # the user is actually subscribing 
        if event == 'SUBSCRIPTION_UPDATE':
            return True


class SMSSendHandler(SMSBaseHandler):
    def get(self, service_type):
        self.respond(service_type)
    def post(self, service_type):
        self.respond(service_type)

    def respond(self, service_type):
        # set the service type
        self.set_service_type(service_type)
        
        # load the parameters
        mobile_number = self.request.get('mobile_number')
        body = self.request.get('body')

        # prepare our response variables
        response = ""
        status = ""

        # check to ensure the designated user is authenticated
        if self.service.is_authenticated(mobile_number):

            # send the message
            self.service.send(mobile_number, body)
            status = "success"
            response = "Sending request to %s" % mobile_number

        # not authenticated, don't send anything
        else:
            status = "fail"
            response = "Not Authenticate: %s" % mobile_number

        logging.info("API response: %s - %s" % (status, response))

        # send the JSON response object
        self.response.headers['Content-Type'] = 'text/plain'
        self.response.out.write(json.dumps({'status': status, 
                                            'mesg': response}))

def main():
    application = webapp.WSGIApplication([(r'/sms/incoming/via/([a-z]+)/?', 
                                           SMSIncomingHandler),
                                          (r'/sms/send/via/([a-z]+)/?', 
                                           SMSSendHandler),
                                          ],
                                         debug=True)
    wsgiref.handlers.CGIHandler().run(application)

if __name__ == '__main__':
    main()
