from geo.geomodel import GeoModel
import geo.geotypes as geotypes
from google.appengine.api import xmpp, users, urlfetch, memcache
from google.appengine.ext import db
from django.utils import simplejson as json
import logging, os, string, random, urllib
from datetime import datetime, timedelta
import math

class Game(object): 
    default_germ_strength = 0.03
    default_germ_speed = 0.0055
    default_germ_spread = 0.03
    default_germ_slowdown = 0.00015
    default_disease_strength = 0.5
    default_disease_strength_increment = 0.001
    default_color_friendly = "#577ff5"
    default_color_enemy = "#007b00"
    default_color_threat = "#ff0000"
    default_score = 5



class CommandCenter(object):
    cc_memcache_key = 'cc-germs'

    def get_germs(self):
        germs = memcache.get(self.cc_memcache_key)
        if germs is None:
            germs = {}

        return germs

    # this will update if it's already in there
    def add_germ(self, germ):
        germs = self.get_germs()
        k = str(germ.key())
        germs[k] = germ.format_for_json(False)

        if not 'deleted' in germs[k]:
            germs[k]['deleted'] = False

        germs[k]['updated'] = GameUtils().date_rfcformat(datetime.now())
        memcache.set(self.cc_memcache_key, germs)
    
    def delete_germ(self, germ):
        germs = self.get_germs()        
        k = str(germ.key())
        if k in germs:
            germs[k]['deleted'] = True
            germs[k]['updated'] = GameUtils().date_rfcformat(datetime.now())

        memcache.set(self.cc_memcache_key, germs)

    def update(self):
        url = "http://pubsubhubbub.appspot.com"
        data = urllib.urlencode({'hub.url': 'http://www.ipandemic.com/feed', 
                                 'hub.mode': 'publish'})
        urlfetch.fetch(url=url, payload=data, method=urlfetch.POST)


class GameUtils(object): 
    # based on: http://www.movable-type.co.uk/scripts/latlong.html
    def distance_in_miles(self, lat1, lon1, lat2, lon2):

        r = 3963 # radius of the earth (in miles)
        # r = 6371 # km 

        lat1 = float(lat1)
        lon1 = float(lon1)
        lat2 = float(lat2)
        lon2 = float(lon2)

        dLat = math.radians(lat2-lat1)
        dLon = math.radians(lon2-lon1)
        a = math.sin(dLat/2) * math.sin(dLat/2) + \
            math.cos(math.radians(lat1)) * math.cos(math.radians(lat2)) * \
            math.sin(dLon/2) * math.sin(dLon/2)
        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1-a)); 
        d = r * c;
        return d

    # http://bugs.python.org/issue7584
    def date_rfcformat(self, dt):
        if dt.tzinfo is None:
            suffix = "-00:00"
        else:
            suffix = dt.strftime("%z")
            suffix = suffix[:-2] + ":" + suffix[-2:]
        return dt.strftime("%Y-%m-%dT%H:%M:%S") + suffix


class UserInfo(GeoModel):
    user = db.UserProperty()
    timestamp = db.DateTimeProperty(auto_now_add=True)
    last_check_in = db.DateTimeProperty()

    score = db.IntegerProperty(default=0)

    # communication
    mobile_key = db.IntegerProperty()
    mobile_number = db.PhoneNumberProperty()
    allow_sms = db.BooleanProperty(default=False)
    allow_xmpp = db.BooleanProperty(default=True)
    
    # game stuff
    default_germ_strength = db.FloatProperty(Game().default_germ_strength)

    def sync(self, current_user):
        # if we have a current_user, ensure that we've logged them to our datastore
        if current_user:
            if not UserInfo().by_user(current_user):

                ui = UserInfo(user=current_user, 
                              mobile_key=random.randint(1000, 9999),
                              score = Game().default_score)
                ui.put()
                xmpp.send_invite(current_user.email())
        else:
            raise Exception("No current user to sync")

    def update_score(self, amount):
        self.score += amount
        self.put()

    def by_user(self, user):
        q = db.GqlQuery("SELECT * FROM UserInfo WHERE user = :1", user)
        u = q.get()
        return u
        
    def by_current_user(self):
        return self.by_user(users.get_current_user())

    def by_mobile_number(self, mobile_number):
        q = db.GqlQuery("SELECT * FROM UserInfo WHERE mobile_number = :1", mobile_number)
        u = q.get()
        return u

    def by_mobile_key(self, mobile_key):
        q = db.GqlQuery("SELECT * FROM UserInfo WHERE mobile_key = :1", int(mobile_key))
        u = q.get()
        return u

    def check_in(self, lat, lon):
        self.location = db.GeoPt(lat, lon);
        self.check_in_time = datetime.now()
        self.update_location()
        self.put()



class Disease(db.Model):
    timestamp = db.DateTimeProperty(auto_now_add=True)
    patient_zero = db.UserProperty()
    current_host = db.UserProperty()
    active = db.BooleanProperty(default=False)
    strain = db.StringProperty() # auto generated name
    name = db.StringProperty()   # let the owner pick the name

    strength = db.FloatProperty(default=Game().default_disease_strength)

    def get_default_disease_for_user(self, user):
        q = Disease().gql("WHERE patient_zero = :1", user)
        disease = q.get()
        if not disease:
            disease = Disease().generate_disease(user)
            disease.put()
        return disease

    def unique_name(self):
        while True:
            s = string.uppercase[random.randint(0, 25)] + \
                str(random.randint(1, 9)) + \
                string.uppercase[random.randint(0, 25)] + \
                str(random.randint(10, 99))
            q = db.GqlQuery("SELECT * FROM Disease WHERE strain = :1", s)
            if not q.count():
                break
        return s

    def generate_disease(self, user):
        disease = Disease()
        disease.patient_zero = user
        disease.current_host = user
#        d.initial_location = db.GeoPt(lat, lon)
        disease.strain = Disease().unique_name()
        return disease



class Germ(GeoModel):
    user = db.UserProperty()
    disease = db.ReferenceProperty(Disease) 
    timestamp = db.DateTimeProperty(auto_now_add=True)

    # how strong is the germ
    strength = db.FloatProperty(default=Game().default_germ_strength)
    
    # the current spead at which it's growing
    spread = db.FloatProperty(default=Game().default_germ_spread)

    # how fast does it increase its spread
    spread_speed = db.FloatProperty(default=Game().default_germ_speed)

    def get_germs_near_point(self, lat, lon, max_distance_in_meters=100):
        center = geotypes.Point(lat, lon)
        base_query = Germ.all()
        data = Germ.proximity_fetch(
            base_query,
            center, max_results=10, max_distance=max_distance_in_meters)
        return data
        

    def fight_nearby_germs(self):
        # get the germs nearby
        spread_in_meters = self.spread * 1609
        germs = Germ().get_germs_near_point(self.location.lat, self.location.lon,
                                            max_distance_in_meters=spread_in_meters)
        for g in germs:
            if g.key() != self.key():
                if self.would_defeat(g):
                    CommandCenter().delete_germ(g)
                    g.delete()


    def auto_spread(self):
        germs = Germ().all()
        strength_increment = Game().default_disease_strength_increment
        for g in germs:
            g.spread += g.spread_speed + 0.001

            if random.randint(0,1) and g.strength > strength_increment:
                g.strength -= strength_increment
                # adjust the strength of the disease
                g.disease.strength -= strength_increment
                g.disease.put()

            if g.spread_speed > Game().default_germ_slowdown:
                g.spread_speed -= Game().default_germ_slowdown

            CommandCenter().add_germ(g)
            g.put()

        CommandCenter().update()            

        
    def auto_fight(self):

        germs = Germ().all()

        # keep track of who needs to go
        losers = []

        # loop through each germ
        for g in germs:
            # no need to check the losers
            if g in losers: continue

            # test it against all the othr germs
            for x in germs:

                # no need to check the losers
                if x in losers: continue


                if x.key() != g.key():

#                    logging.info("FIGHT fighting: %s", x.user)

                    distance = GameUtils().distance_in_miles(g.location.lat,
                                                            g.location.lon,
                                                            x.location.lat,
                                                            x.location.lon)

                    # are they close together? They should fight
                    if distance < g.spread:
                        if g.would_defeat(x):
                            losers.append(x)
                        else:
                            losers.append(g)


        for loser in losers:
            # send a message to the user...
            CommandCenter().delete_germ(loser)
            loser.delete()

        if len(losers):
            logging.info("there are some losers, update command center")
            CommandCenter().update()
            

    def invigorate(self):
        # fixme: this should only happen if you haven't check in here in a while
        # until then, how's a bit of randomness sound?
        if random.randint(1,9) != 7:
            logging.info("INVIGORATING!, guessed the right number")
            self.strength = self.strength * 1.5
            self.spread_speed = Game().default_germ_speed
            self.put()
        
            logging.info("increasing disease strength by %f" % (self.strength / 1.5))
            self.disease.strength += (self.strength / 1.5)
            self.disease.put()

            CommandCenter().add_germ(self)
            CommandCenter().update()
        else:
            logging.info("NOT INVIGORATING!")


    def would_defeat(self, enemy):
        
        our_strength = self.disease.strength + self.strength
        enemy_strength = enemy.disease.strength + enemy.strength
        
        logging.info("our_strength: %f, enemy_strength: %f" % (our_strength, enemy_strength))

        if our_strength >= enemy_strength:
            logging.info("Germ 1 (our_strength) wins!")
            return True
        else:
#            logging.info("ENEMY wins!")
            return False
            
        
    def format_for_json(self, current_user):
        return {
            'id': str(self.key()),
            'lat': self.location.lat,
            'lon': self.location.lon,
            'strength': self.strength,
            'spread': self.spread,
            'timestamp': GameUtils().date_rfcformat(self.timestamp),
            'threat': False,
            'disease': { 
                'id': str(self.disease.key()),
                'strain': self.disease.strain,
                'name': self.disease.name,
                'host': self.disease.current_host.nickname(),
                'host_is_you': self.disease.current_host == current_user,
                }
            }


class UserThreats(db.Model):
    user = db.UserProperty()
    response_key = db.IntegerProperty()
    threat_type = db.StringProperty()    
    expire_time = db.DateTimeProperty()
    germ_threat = db.ReferenceProperty(Germ)
    challenge_user = db.UserProperty()
    
    def create(self, user, germ_threat, expires_in_minutes=5, challenge_user=None, threat_type="natural"):
        q = UserThreats.all()
        q.filter('user = ', user)
        q.filter('challenge_user = ', challenge_user)
        q.filter('germ_threat = ', germ_threat)
        q.filter('threat_type =', threat_type)
        if not q.count():
            expire_time = (timedelta(minutes=expires_in_minutes) + datetime.now())
            timeformat = '%Y-%m-%d %H:%M:%S'
            logging.info("CREATING THE UserThreats %s!", expire_time)
            logging.info(expire_time)

            ut = UserThreats(user = user,
                             challenge_user = challenge_user,
                             threat_type = threat_type,
                             germ_threat = germ_threat,
                             expire_time = expire_time)

            # if this is a challenge threat, setup a new response key
            if threat_type == 'challenge':
                # get the biggest response key
                cq = UserThreats.all()
                cq.filter('user = ', user)
                cq.order('-response_key')
                biggest = cq.get()
                if biggest and biggest.response_key:
                    ut.response_key = biggest.response_key + 1
                else:
                    ut.response_key = 1

            ut.put()
            return ut
        

    def respond_to_threat(self, user, response_key, response):
        # load the response using the user/response key combo
        q = UserThreats.all()
        q.filter('user = ', user)
        q.filter('response_key = ', response_key)
        threat = q.get()
      
        # this method returns the actual response that gets sent
        # to the user.
        response_to_sender = ""

        # did they respond to a challenge threat?
        # let's increase the strength of the challenged germ
        if threat.threat_type == 'challenge':
            challenge_user = threat.challenge_user
            threat.germ_threat.strength += Game().default_disease_strength_increment

            UserInfo().by_user(user).update_score(3)
            UserInfo().by_user(challenge_user).update_score(-1)

            response_to_sender = "Awesome! You defeated a challenge! Points, strength, good looks? All increased"

            # forward the message to the loser of the challenge
            message = "%s: %s\n\nYou lost the challenge! You should probably cover up." % (user.nickname(), response)
            logging.info("SENDING %s" % response)
            Messenger().send(user, challenge_user, message)


        # The threat has been responded to!
        # we can delte it
        threat.delete()
        
        return response_to_sender

    def remove_threat(self, user, germ):
        q = UserThreats.all()
        q.filter('user = ', user)
        q.filter('germ_threat = ', germ)
        threat = q.get()
        if threat:
            logging.info("removing the threat")
            threat.delete()
        else:
            logging.info("no threat to delete... ignoring")

    def update_threats(self):
        timeformat = '%Y-%m-%d %H:%M:%S'
        logging.info(datetime.now().strftime(timeformat))
        q = UserThreats.gql('WHERE expire_time < datetime(:1)', datetime.now().strftime(timeformat))
        
        for x in q:
            logging.info("GOT AN EXPIRED THREAT! %s, %s %s" % (x.user.email(), x.threat_type, x.expire_time))

            if x.threat_type == 'challenge':

                UserInfo().by_user(x.user).update_score(-3)
                message = "Someone threatened you and you didn't respond. Points deducted! Also, your germs died!"
                Messenger().send(None, x.user, message)
                
                try:
                    # # the other user gets points for scaring someone off!
                    UserInfo().by_user(x.challenge_user).update_score(+3)
                except Exception, e:
                    logging.error("Unable to update score for challenge user:")
                    logging.error(e)

                germ_threat = x.germ_threat
                        
                try:                        
                    # update the overall strength of the disease
                    germ_threat.disease.strength -= germ_threat.strength
                    germ_threat.disease.put()
                except Exception, e:
                    logging.error("Unable to update disease strength for germ threat:")
                    logging.error(e)
                        
                try:
                    tq = UserThreats.all()
                    tq.filter('user = ', x.challenge_user)
                    tq.filter('germ_threat = ', germ_threat)
                    tq.filter('threat_type = ', 'natural')
                    natural_threat = tq.get()
                    if natural_threat:
                        logging.info("DELETING THE NATURAL THREAT AFTER A CHALLENGE")
                        natural_threat.delete()
                    else:
                        logging.info("No natural threat?")
                        natural_threat.delete()
                except Exception, e:
                    logging.error("problem deleting the natural threat?")
                    logging.error(e)

                try:
                    # delete the germ that caused the threat
                    CommandCenter().delete_germ(germ_threat)
                    CommandCenter().update()
                    germ_threat.delete()
                except Exception, e:
                    logging.error("Unable to update the command center")
                    logging.error(e)

                message = "Good challenge. You wiped out the other disease. (Refresh and) cough again to spread yours!"
                Messenger().send(None, x.challenge_user, message)
                    

            elif x.threat_type == 'natural':

                logging.info("GOT AN EXPIRED NaturalThreat! %s" % x.user.email())

                try:
                    # this user didn't do anything!
                    UserInfo().by_user(x.user).update_score(-3)

                    # the other user scared someone off
                    UserInfo().by_user(x.germ_threat.user).update_score(+3)

                except Exception, e:
                    logging.error("EXCEPTION updating scores: %s" % e)
                    pass

                logging.info("from update-threats Sending message to %s" % x.user)
                message = "You were threatened by a disease on your last check-in and didn't cough or challenge or anything. Points deducted!"
                Messenger().send(None, x.user, message)
                logging.info("Returned from message thing")

            # delete this threat either way
            x.delete()

class Textmark(object):
    auth_user = '6463300181'  # USER DEFINED
    auth_pass = '7075'  # USER DEFINED
    short_code = '41411'
    keyword = 'sick'
    api_key = 'tedroden_com_884c8f59'  # USER DEFINEDapi_key


    # override the send method from SMSService
    def send(self, number, body):
        # The Send URL
        url = "http://dev1.api2.textmarks.com/GroupLeader/send_one_message/"

        # setup the http arguments
        args = {}
        args['to'] = number             
        args['msg'] = body
        args['tm'] = self.keyword
        args['api_key'] = self.api_key       
        args['auth_user'] = self.auth_user
        args['auth_pass'] = self.auth_pass   

        # make the HTTP API call to send the message
        result = urlfetch.fetch(url=url,
                                payload=urllib.urlencode(args),
                                method=urlfetch.POST)
        logging.info(result.content)



class Messenger(object):
    def send_via_sms(self, mobile_number, body):
        sms = Textmark()
        sms.send(mobile_number, body)

    def send_via_xmpp(self, to, body):
        body += "\n\nTo stop receiving message reply with one word: stop\n"
        xmpp_response = xmpp.send_message(to, body)
        logging.info(xmpp_response)
        
        
    def send(self, from_user, to_user, body, response_key=False):
        if from_user:
            body = from_user.nickname() + ": " + body

        logging.info("SEND?: " + body)
        ui = UserInfo().by_user(to_user)
        if ui.allow_xmpp:
            im_body = body
            if response_key:
                im_body += " Respond with @%d [your message here]" % int(response_key)
            logging.info("Allowed to snd IM")
            self.send_via_xmpp(to_user.email(), im_body)

        if ui.mobile_number and ui.allow_sms:
            sms_body = body
            if response_key:
                sms_body += " Respond with \"%s @%d [your message here]\"" % (Textmark.keyword, int(response_key))
            logging.info("Allowed to send SMS")
            self.send_via_sms(ui.mobile_number, sms_body)
        
        

