# -*- indent-tabs-mode: nil -*-
import os, logging
import tornado.httpserver
import tornado.ioloop
import tornado.web
import tornado.auth
import simplejson as json
import time, re, uuid
from tornado.options import define, options
from tornado.escape import url_escape
import urllib, urllib2
from xml.dom import minidom

# Define options that can be changed as we run this via the command line
define("port", default=8088, help="Run server on a specific port", type=int)
define("use_phsb", default=True, help="Use PubSubHubbub?", type=bool)

class Notifier(object):
    listeners = []
    germs = []
    message_cache = []        # a list of recent chat messages
    message_cache_size = 20   # the amount of recent messages to store

    def add_listener(self, callback, cursor=None):
        # if they've supplied a cursor, send them any messages since that cursor
        if cursor:
            index = 0
            for i in xrange(len(self.message_cache)): 
                index = len(self.message_cache) - i - 1
                if self.message_cache[index]["id"] == cursor: break 
            recent = self.message_cache[index + 1:]
            if recent:
                callback(messages=recent)
                return

        # if they haven't sent a cursor, send them everything in the cache 
        if not cursor and len(self.message_cache):
            callback(messages=self.message_cache, germs=self.germs)
            return
        
        # if they've made it this far, we have nothing to show them. wait it out
        self.listeners.append(callback)

    def notify_listeners(self, messages=False, germs=False):
        # call the callback method on each of the listeners
        for callback in self.listeners:
            try:
                callback(messages=messages, germs=germs)
            except:
                logging.error("Error in listeners callback", exc_info=True)

        # clear the listeners (we've just sent something to each of them)
        self.listeners = []


    def send_message(self, text):
        message = {'text': text,
                   'id': str(uuid.uuid4()) }

        self.notify_listeners(messages=[message])

        # and update the message cache
        self.message_cache.extend([message])
        if len(self.message_cache) > self.message_cache_size:
            self.message_cache = self.message_cache[-self.message_cache_size:]

    def add_germ(self, germ):
        x = 0
        # if we already have this germ, update the data and return
        while x < len(self.germs):
            if self.germs[x]['id'] == germ['id']:
                self.germs[x] = germ
                return
            x += 1

        # otherwise, just append the germ
        self.germs.append(germ)

    def update_clients(self):
        # notify everyone about the germs
        self.notify_listeners(germs=self.germs)

        
class GoogleLoginHandler(tornado.web.RequestHandler, tornado.auth.GoogleMixin):
    @tornado.web.asynchronous
    def get(self):
        if self.get_argument("openid.mode", None):
            self.get_authenticated_user(self.async_callback(self._on_auth))
            return
        self.authenticate_redirect()

    def _on_auth(self, user):
        if not user:
            self.authenticate_redirect()
            return
        else:
            self.set_secure_cookie('user', json.dumps(user))
        self.application.notifier.send_message(user['first_name'] + " joined the chat")
        self.redirect("/")

class LogoutHandler(tornado.web.RequestHandler): 
    def get(self):
        self.clear_cookie("user")
        self.redirect('/')


class HomeHandler(tornado.web.RequestHandler):
    def get(self):
        user_json = self.get_secure_cookie('user')
        user = False
        if user_json: user = json.loads(user_json)

        self.render("control-center.html", 
                    user=user)

class ChatSayHandler(tornado.web.RequestHandler):

    def post(self):
        user_json = self.get_secure_cookie('user')
        if user_json:
            user = json.loads(user_json)
        
            text = self.get_argument('text').strip()
            logging.info("TEXT: %s", text)
            if len(text):
                text = "<strong>%s</strong>: %s" % (user['first_name'], text)
                self.application.notifier.send_message(text)

        self.finish({'status': 'ok!'})

# wait for updates and return them when available
class UpdateHandler(tornado.web.RequestHandler):

    @tornado.web.asynchronous
    def post(self):
        cursor = self.get_argument('chat_cursor', False)
        self.application.notifier.add_listener(self.async_callback(self.handle_updates), cursor)

    def handle_updates(self, messages=False, germs=False): 
        if not self.request.connection.stream.closed():
            self.finish({'chat_messages': messages, 'germs': germs})



# wait for updates and return them when available
class CatchGermsHandler(tornado.web.RequestHandler):
    def get(self):
        if self.get_argument('hub.mode', False):
            logging.info("responding to hub challenge")
            self.finish(self.get_argument('hub.challenge'))

        # let's also use this URL to get updates from the server
        else:
            logging.info("Just pinging the server!")
            response = urllib2.urlopen(self.application.topic_url)
            self.parse_feed(response.read())
            self.finish({'status': 'ok!'})
            
    def post(self):
        self.parse_feed(self.request.body)

    def parse_feed(self, feed_body):
        dom = minidom.parseString(feed_body)
        ATOM_NS = "http://www.w3.org/2005/Atom"

        for node in dom.getElementsByTagNameNS(ATOM_NS, 'content'):
            germ = json.loads(node.firstChild.data.strip())
            self.application.notifier.add_germ(germ)
            
        self.application.notifier.update_clients()

class Application(tornado.web.Application):
    def __init__(self):

        handlers = [
            (r"/", HomeHandler),
            (r"/chat/say/?", ChatSayHandler),
            (r"/catch/germs/?", CatchGermsHandler),
            (r"/updates/?", UpdateHandler),
            (r"/login/?", GoogleLoginHandler),
            (r"/logout/?", LogoutHandler),
            ]

        settings = {
            "static_path": os.path.join(os.path.dirname(__file__), "static"),
            "template_path": os.path.join(os.path.dirname(__file__), "templates"),
            "cookie_secret": "4$983c98%^eclk238cLJadfec3",
            "login_url": "/login",
            }
        
        tornado.web.Application.__init__(self, handlers, **settings)
        self.notifier = Notifier()
        self.app_host = 'http://cdc.ipandemic.com:%d' % options.port
        self.topic_url = "http://www.ipandemic.com/feed"
        self.hub_url = "http://pubsubhubbub.appspot.com"
        self.callback_url = self.app_host + "/catch/germs"

        if options.use_phsb:
            # set up the different urls we need to know
            logging.info("Subscribing to the HUB")
            data = urllib.urlencode({'hub.topic': self.topic_url,
                                     'hub.callback': self.callback_url,
                                     'hub.verify': 'async',
                                     'hub.mode': 'subscribe'})
    
            try:
                response = urllib2.urlopen(self.hub_url, data)
            except (IOError, urllib2.HTTPError), e:
                if hasattr(e, 'code') and e.code == 204:
                    logging.info("204: No Content")
                error = ''
                if hasattr(e, 'read'):
                    error = e.read()
                logging.info('%s, Response: "%s"' % (e, error))
        
        else:
            logging.info("Not using PubSubHubbub, call /catch/germs manually.")
            

        
        

if __name__ == "__main__":
    tornado.options.parse_command_line()
    http_server = tornado.httpserver.HTTPServer(Application())
    http_server.listen(options.port)
    tornado.ioloop.IOLoop.instance().start()
